@@ -59,6 +59,10 @@ function zend_string_or_stdclass_or_null($param): stdClass|string|null {}
 
 function zend_iterable(iterable $arg1, ?iterable $arg2 = null): void {}
 
+function zend_weakmap_attach(object $object, mixed $value): bool {}
+function zend_weakmap_remove(object $object): bool {}
+function zend_weakmap_dump(): array {}
+
 }
 
 namespace ZendTestNS {
